from __future__ import annotations

import itertools
import random
from collections import Counter

import torch
from qadence import NoiseHandler, QuantumCircuit, QuantumModel, X, chain, kron
from qadence.blocks.utils import unroll_block_with_scaling
from qadence.measurements.samples import compute_expectation
from torch import Tensor, tensor


def twirl_swap(n_qubits: int, twirl: tuple, samples_twirl: Counter[str]) -> dict:
    """
    Applies the corresponding twirl operations (multi-qubit bit flip).

    Achieved by remapping the measurements appropriately
    """
    output = {}
    operand = "".join(map(str, [1 if i in twirl else 0 for i in range(n_qubits)]))
    for key in samples_twirl.keys():
        output[bin(int(key, 2) ^ int(operand, 2))[2:].zfill(n_qubits)] = samples_twirl[key]

    return Counter(output)


def mitigate(
    model: QuantumModel,
    options: dict,
    noise: NoiseHandler | None = None,
    param_values: dict[str, Tensor] = dict(),
) -> tensor:
    """Corrects for readout errors on expectation values using all possible twirl operations.

    See Equation at the end of page 3
    See Page(2) in https://arxiv.org/pdf/2012.09738.pdf for implementation

    Args:
        model: model with a specified readout error whose samples need to be mitogated
        options: specify any additional parameters that the protcol might require

    Returns:
        Mitigated output is returned
    """
    block = model._circuit.original.block
    n_shots = model._measurement.options["n_shots"]

    twirl_samples = options.get("twirl_samples", None)
    all_qubits = list(range(block.n_qubits))
    num_total_comb = (2**block.n_qubits) - 1

    # Validity check for twirl_samples
    if twirl_samples is not None:
        if not (isinstance(twirl_samples, int) and 1 <= twirl_samples <= num_total_comb):
            raise ValueError(
                f"twirl_samples must be an integer type between 1 and {num_total_comb}"
            )

        # If twirl_samples is given, generate samples with the given value
        sampled_twirls: set[tuple[int, ...]] = set()
        while len(sampled_twirls) < twirl_samples:
            k = random.randint(1, block.n_qubits)  # Random size of the combination
            sampled_twirls.add(tuple(random.sample(all_qubits, k)))
        twirls = list(sampled_twirls)

    # If twirl_samples is None, generate all combinations
    else:
        twirls = sum(
            (list(itertools.combinations(all_qubits, k)) for k in range(1, block.n_qubits + 1)),
            [],
        )

    # Generate samples for all twirls of circuit
    samples_twirl_num_list = []
    samples_twirl_den_list = []
    for twirl in twirls:
        block_twirl = chain(block, kron(X(i) for i in twirl))

        # Twirl outputs for given circuit (Numerator)
        circ_twirl_num = QuantumCircuit(block_twirl.n_qubits, block_twirl)
        model_twirl_num = QuantumModel(circuit=circ_twirl_num, backend=model._backend_name)
        samples_twirl_num = model_twirl_num.sample(noise=model._noise, n_shots=n_shots)[0]
        samples_twirl_num_list.append(twirl_swap(block_twirl.n_qubits, twirl, samples_twirl_num))

        # Twirl outputs on input state (Denominator)
        circ_twirl_den = QuantumCircuit(block_twirl.n_qubits, kron(X(i) for i in twirl))
        model_twirl_den = QuantumModel(circuit=circ_twirl_den, backend=model._backend_name)
        samples_twirl_den = model_twirl_den.sample(noise=model._noise, n_shots=n_shots)[0]
        samples_twirl_den_list.append(twirl_swap(block_twirl.n_qubits, twirl, samples_twirl_den))

    output_exp = []

    for observable in model._observable:
        expectation: float = 0

        coeffs = torch.tensor(
            [pauli[1] for pauli in unroll_block_with_scaling(observable.original)],
            dtype=float,
        )
        obs_list = [pauli[0] for pauli in unroll_block_with_scaling(observable.original)]

        expectation_num = torch.stack(compute_expectation(obs_list, samples_twirl_num_list))
        expectation_den = torch.stack(compute_expectation(obs_list, samples_twirl_den_list))

        expectation = torch.sum(expectation_num, dim=1) / torch.sum(expectation_den, dim=1)

        output_exp.append(torch.sum(torch.dot(coeffs, expectation)))

    return tensor(output_exp)



from __future__ import annotations

from typing import Callable

import numpy as np
import torch
from qadence import BackendName, QuantumModel
from qadence.backends.api import backend_factory
from qadence.backends.pulser.backend import Backend
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.analog import ConstantAnalogRotation, InteractionBlock
from qadence.circuit import QuantumCircuit
from qadence.noise import NoiseHandler
from qadence.operations import AnalogRot
from qadence.transpile import apply_fn_to_blocks
from qadence.types import NoiseProtocol
from qadence.utils import Endianness
from scipy.optimize import curve_fit
from torch import Tensor

supported_noise_models = [
    NoiseProtocol.ANALOG.DEPOLARIZING,
    NoiseProtocol.ANALOG.DEPHASING,
]


def zne_poly(noise_levels: Tensor, zne_datasets: list[list]) -> Tensor:
    poly_fits = []

    # check if datapoints are enough
    if len(noise_levels) < 2:
        raise ValueError("At least 2 noise levels are required for polynomial fitting.")

    for dataset in zne_datasets:  # Looping over batched observables.
        poly_fit = np.poly1d(np.polyfit(noise_levels, dataset, len(noise_levels) - 1))
        # Return the zero-noise extrapolated value.
        poly_fits.append(poly_fit(0.0))
    return torch.tensor(poly_fits)


def zne_exp(noise_levels: Tensor, zne_datasets: list[list[float]]) -> Tensor:
    def exp_fn(x: np.ndarray, a: float, b: float, c: float) -> np.ndarray:
        return a * np.exp(-b * x) + c

    exp_fits: list[float] = []
    noise_levels_np: np.ndarray = noise_levels.numpy()

    for dataset in zne_datasets:
        dataset_np: np.ndarray = np.array(dataset)

        # check if datapoints are enough
        if len(noise_levels_np) < 3:
            raise ValueError("At least 3 noise levels are required for exponential fitting.")

        try:
            # Execute fitting
            popt, _ = curve_fit(
                exp_fn,
                noise_levels_np,
                dataset_np,
                p0=(dataset_np[0], 1, dataset_np[-1]),
                maxfev=5000,
            )

            # expected value when noise is zero
            zero_noise_value: float = float(exp_fn(0, *popt))
            exp_fits.append(zero_noise_value)

        except RuntimeError:
            raise ValueError(
                "Optimal parameters not found, try with other datapoints or fitting configuration."
            )

    return torch.tensor(exp_fits)


def pulse_experiment(
    backend: Backend,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor],
    noise: NoiseHandler,
    stretches: Tensor,
    endianness: Endianness,
    zne_func: Callable,
    state: Tensor | None = None,
) -> Tensor:
    def mutate_params(block: AbstractBlock, stretch: float) -> AbstractBlock:
        """Closure to retrieve and stretch analog parameters."""
        # Check for stretchable analog block.
        if isinstance(block, (ConstantAnalogRotation, InteractionBlock)):
            stretched_duration = block.parameters.duration * stretch
            stretched_omega = block.parameters.omega / stretch
            stretched_delta = block.parameters.delta / stretch
            # The Hamiltonian scaling has no effect on the phase parameter
            phase = block.parameters.phase
            qubit_support = block.qubit_support
            return AnalogRot(
                duration=stretched_duration,
                omega=stretched_omega,
                delta=stretched_delta,
                phase=phase,
                qubit_support=qubit_support,
            )
        return block

    converted_observables = [backend.observable(obs) for obs in observable]
    noisy_density_matrices: list = []
    for stretch in stretches:
        # FIXME: Iterating through the circuit for every stretch
        # and rebuilding the block leaves is inefficient.
        # Best to retrieve the parameters once
        # and rebuild the blocks.
        stre = stretch.item()
        block = apply_fn_to_blocks(circuit.block, mutate_params, stre)
        stretched_register = circuit.register.rescale_coords(stre)
        stretched_circuit = QuantumCircuit(stretched_register, block)
        converted_circuit = backend.circuit(stretched_circuit)
        noisy_density_matrices.append(
            # Contain a single experiment result for the stretch.
            backend.run(
                converted_circuit,
                param_values=param_values,
                state=state,
                noise=noise,
                endianness=endianness,
            )[0]
        )
    support = sorted(list(circuit.register.support))
    # It is not possible to use the expectation method as the pulse stretching concerns the circuit.
    res_list = [
        [
            obs.native(dm, param_values, qubit_support=support, noise=noise)
            for dm in noisy_density_matrices
        ]
        for obs in converted_observables
    ]
    res = torch.stack([torch.transpose(torch.stack(res), 0, -1) for res in res_list])
    res = res if len(res.shape) > 0 else res.reshape(1)

    # Zero-noise extrapolate.
    extrapolated_exp_values = zne_func(
        noise_levels=stretches,
        zne_datasets=res.real,
    )
    return extrapolated_exp_values


def noise_level_experiment(
    backend: Backend,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor],
    noise: NoiseHandler,
    endianness: Endianness,
    zne_func: Callable,
    state: Tensor | None = None,
) -> Tensor:
    noise_probs = noise.options[-1].get("noise_probs")
    converted_circuit = backend.circuit(circuit)
    converted_observables = [backend.observable(obs) for obs in observable]
    zne_datasets = backend.expectation(
        converted_circuit,
        converted_observables,
        param_values=param_values,
        state=state,
        noise=noise,
        endianness=endianness,
    )

    return zne_func(noise_levels=noise_probs, zne_datasets=zne_datasets)


def analog_zne(
    model: QuantumModel,
    options: dict,
    noise: NoiseHandler,
    param_values: dict[str, Tensor],
    state: Tensor | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    if model._backend_name != BackendName.PULSER:
        raise ValueError("Only BackendName.PULSER supports analog simulations.")
    backend = backend_factory(backend=model._backend_name, diff_mode=None)
    stretches = options.get("stretches", None)
    zne_type = options.get("zne_type", "poly")

    if zne_type == "poly":
        zne_func = zne_poly
    elif zne_type == "exp":
        zne_func = zne_exp
    else:
        raise ValueError(
            f'Analog ZNE supports only polynomial ("poly") or exponential ("exp") '
            f"extrapolation options. Got {zne_type}."
        )

    if stretches is not None:
        extrapolated_exp_values = pulse_experiment(
            backend=backend,
            circuit=model._circuit.original,
            observable=[obs.original for obs in model._observable],
            param_values=param_values,
            noise=noise,
            stretches=stretches,
            endianness=endianness,
            zne_func=zne_func,
            state=state,
        )
    else:
        extrapolated_exp_values = noise_level_experiment(
            backend=backend,
            circuit=model._circuit.original,
            observable=[obs.original for obs in model._observable],
            param_values=param_values,
            noise=noise,
            endianness=endianness,
            zne_func=zne_func,
            state=state,
        )
    return extrapolated_exp_values


def mitigate(
    model: QuantumModel,
    options: dict,
    noise: NoiseHandler,
    param_values: dict[str, Tensor] = dict(),
) -> Tensor:
    if noise.protocol[-1] not in supported_noise_models:
        raise ValueError("A NoiseProtocol.ANALOG noise model must be provided to .mitigate()")
    mitigation_zne = analog_zne(
        model=model, options=options, noise=noise, param_values=param_values
    )
    return mitigation_zne



from __future__ import annotations

from collections import Counter
from itertools import combinations

import numpy as np
import numpy.typing as npt
from numpy.linalg import inv, matrix_rank, pinv
from pyqtorch.noise import CorrelatedReadoutNoise
from qadence import QuantumModel
from qadence.backends.pyqtorch.convert_ops import convert_readout_noise
from qadence.logger import get_logger
from qadence.noise.protocols import NoiseHandler
from qadence.types import NoiseProtocol
from scipy.linalg import norm
from scipy.optimize import LinearConstraint, minimize
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import gmres
from torch import Tensor

from qadence_mitigation.types import ReadOutOptimization

logger = get_logger(__name__)


def normalized_subspace_kron(
    noise_matrices: npt.NDArray, subspace: npt.NDArray, hamming_dist: int | None = None
) -> npt.NDArray:
    """
    Compute a specified tensor-producted subspace of index locations.

    Args:
        noise_matrices: Array of noise matrices acting on independent qubits.
        subspace: List of index locations defining the subspace for computation.
        hamming_dist: Optional Hamming distance threshold for including nearby indices.

    Returns:
        A sparse matrix constructed from the tensor product of noise matrices in the subspace.
    """

    n_qubits = len(noise_matrices)
    conf_matrix = csr_matrix((2**n_qubits, 2**n_qubits))

    # Expand subspace based on Hamming distance if specified
    if hamming_dist is not None:
        valid_indices = set(subspace)
        for idx in subspace:
            binary_str = bin(idx)[2:].zfill(n_qubits)

            # Generate all possible indices within hamming_dist
            for dist in range(1, hamming_dist + 1):
                for flip_positions in combinations(range(n_qubits), dist):
                    flipped_binary = list(binary_str)
                    for pos in flip_positions:
                        flipped_binary[pos] = "1" if flipped_binary[pos] == "0" else "0"
                    valid_indices.add(int("".join(flipped_binary), 2))
        subspace = sorted(valid_indices)

    # Compute tensor-producted confusion matrix for the subspace
    for j in subspace:
        for i in subspace:
            bin_i, bin_j = bin(i)[2:].zfill(n_qubits), bin(j)[2:].zfill(n_qubits)

            # Manually computing the entries of tensor product for only the subspace
            conf_matrix[i, j] = np.prod(
                [noise_matrices[k][int(bin_i[k])][int(bin_j[k])] for k in range(n_qubits)]
            )
        conf_matrix[:, j] /= np.sum(conf_matrix[:, j])

    return conf_matrix


def tensor_rank_mult(qubit_ops: npt.NDArray, prob_vect: npt.NDArray) -> npt.NDArray:
    """
    Fast multiplication of single qubit operators on a probability vector.

    Similar to how gate operations are implemented on PyqTorch
    Needs to be replaced.
    """

    n_qubits = int(np.log2(len(prob_vect)))

    # Reshape probability vector into a rank N tensor
    prob_vect_t = prob_vect.reshape(n_qubits * [2]).transpose()

    # Contract each tensor index (qubit) with the inverse of the single-qubit
    for i in range(n_qubits):
        prob_vect_t = np.tensordot(qubit_ops[n_qubits - 1 - i], prob_vect_t, axes=(1, i))

    # Obtain corrected measurements by shaping back into a vector
    return prob_vect_t.reshape(2**n_qubits)


def corrected_probas(
    p_corr: npt.NDArray, noise_matrices: npt.NDArray, p_raw: npt.NDArray
) -> np.double:
    ## Computing rectified probabilites without computing the full T matrix
    p_estim = tensor_rank_mult(noise_matrices, p_corr.T)
    return norm(p_estim - p_raw.T, ord=2) ** 2


def mle_solve(p_raw: npt.NDArray) -> npt.NDArray:
    """
    Compute the MLE probability vector.

    Algorithmic details can be found in https://arxiv.org/pdf/1106.5458.pdf Page(3).
    """
    # Sort p_raw by values while keeping track of indices.
    index_sort = p_raw.argsort()
    p_sort = p_raw[index_sort]
    neg_sum = 0
    breakpoint = len(p_sort) - 1

    for i in range(len(p_sort)):
        # If neg_sum cannot be distributed among other probabilities, continue to accumulate
        if p_sort[i] + neg_sum / (len(p_sort) - i) < 0:
            neg_sum += p_sort[i]
            p_sort[i] = 0
        # Set breakpoint to current index
        else:
            breakpoint = i
            break
    # Number of entries to which i can distribute(includes breakpoint)
    size = len(p_sort) - breakpoint
    p_sort[breakpoint:] += neg_sum / size

    re_index_sort = index_sort.argsort()
    p_corr = p_sort[re_index_sort]

    return p_corr


def renormalize_counts(corrected_counts: npt.NDArray, n_shots: int) -> npt.NDArray:
    """Renormalize counts rounding discrepancies."""
    total_counts = sum(corrected_counts)
    if total_counts != n_shots:
        counts_diff = total_counts - n_shots
        corrected_counts -= counts_diff
        corrected_counts = np.where(corrected_counts < 0, 0, corrected_counts)
        sum_corrected_counts = sum(corrected_counts)

        renormalization_factor = n_shots / sum_corrected_counts
        corrected_counts = np.rint(corrected_counts * renormalization_factor).astype(int)

    # At this point, the count should be off by at most 2, added or substracted to/from the
    # max count.
    if sum(corrected_counts) != n_shots:
        count_diff = sum(corrected_counts) - n_shots
        max_count_bs = np.argmax(corrected_counts)
        corrected_counts[max_count_bs] -= count_diff

    return corrected_counts


def matrix_inv(K: npt.NDArray) -> npt.NDArray:
    return inv(K) if matrix_rank(K) == K.shape[0] else pinv(K)


def constrained_inversion(noise_matrices: npt.NDArray, p_raw: npt.NDArray) -> npt.NDArray:
    # Initial random guess in [0,1].
    p_corr0 = np.random.rand(len(p_raw))
    # Stochasticity constraints.
    normality_constraint = LinearConstraint(np.ones(len(p_raw)).astype(int), lb=1.0, ub=1.0)
    positivity_constraint = LinearConstraint(np.eye(len(p_raw)).astype(int), lb=0.0, ub=1.0)
    constraints = [normality_constraint, positivity_constraint]
    # Minimize the corrected probabilities.
    res = minimize(corrected_probas, p_corr0, args=(noise_matrices, p_raw), constraints=constraints)

    return res.x


def majority_vote(noise_matrices: npt.NDArray, p_raw: npt.NDArray) -> npt.NDArray:
    """
    Compute the bitstring with the highest probability of occurrrence via voting.

    For implementation, see Equation (6) and Page 4 in https://arxiv.org/pdf/2402.11830.pdf
    """
    n_qubits = len(noise_matrices)
    output = 0

    for i in range(n_qubits):
        p_raw_resize = p_raw.reshape([2] * n_qubits)
        transposed_axes = [i] + list(range(0, i)) + list(range(i + 1, n_qubits))
        p_raw_resize = np.transpose(p_raw_resize, axes=transposed_axes).reshape(2, 2**n_qubits // 2)
        probs = np.sum(p_raw_resize, axis=1)

        # Given the output to be 0, the probability of observed outcomes
        prob_zero = noise_matrices[i][1][0] ** probs[1] * noise_matrices[i][0][0] ** probs[0]
        # Given the output to be 0, the probability of observed outcomes
        prob_one = noise_matrices[i][1][1] ** probs[1] * noise_matrices[i][0][1] ** probs[0]

        if prob_one > prob_zero:
            output += 2 ** (n_qubits - 1 - i)

    p_corr = np.zeros(2**n_qubits)
    p_corr[output] = 1

    return p_corr


def mitigation_minimization(
    noise: NoiseHandler,
    options: dict,
    samples: list[Counter],
) -> list[Counter]:
    """Minimize a correction matrix subjected to stochasticity constraints.

    See Equation (5) in https://arxiv.org/pdf/2001.09980.pdf.
    See Page(3) in https://arxiv.org/pdf/1106.5458.pdf for MLE implementation
    See Equation (5) in https://arxiv.org/pdf/2108.12518.pdf for MTHREE implementation
    (matrix free measurement mitigation)
    This method is reserved for implementations of over 20 qubits
    See Equation (6) and Page 4 in https://arxiv.org/pdf/2402.11830.pdf for MV implementation
    This method is reserved for algorithms with a single bit string output
    See page (5) for extension to more than one solution (inefficient)

    Args:
        noise: Specifies confusion matrix and default error probability
        mitigation: Selects additional mitigation options based on noise choice.
        For readout we have the following mitigation options for optimization
        1. constrained 2. mle (Default) mle 3. mthree 4. majority vote (mv)
        samples: List of samples to be mitigated

    Returns:
        Mitigated counts computed by the algorithm
    """
    n_qubits = len(list(samples[0].keys())[0])
    readout_noise = convert_readout_noise(n_qubits, noise)
    if readout_noise is None or isinstance(readout_noise, CorrelatedReadoutNoise):
        raise ValueError("Specify a noise source of type NoiseProtocol.READOUT.INDEPENDENT.")
    n_shots = sum(samples[0].values())
    noise_matrices = readout_noise.confusion_matrix
    if readout_noise._compute_confusion:
        noise_matrices = readout_noise.create_noise_matrix(n_shots)
        noise_matrices = readout_noise.confusion_matrix
    noise_matrices = noise_matrices.numpy()

    optimization_type = options.get("optimization_type", ReadOutOptimization.MLE)
    corrected_counters: list[Counter] = []

    for sample in samples:
        ordered_bitstrings = [bin(k)[2:].zfill(n_qubits) for k in range(2**n_qubits)]
        # Array of raw probabilites.
        p_raw = np.array([sample[bs] for bs in ordered_bitstrings]) / n_shots

        if optimization_type == ReadOutOptimization.CONSTRAINED:
            p_corr = constrained_inversion(noise_matrices, p_raw)

        elif optimization_type == ReadOutOptimization.MLE:
            noise_matrices_inv = list(map(matrix_inv, noise_matrices))
            # Compute corrected inverse using matrix inversion and run MLE.
            p_corr = mle_solve(tensor_rank_mult(noise_matrices_inv, p_raw))

        elif optimization_type == ReadOutOptimization.MTHREE:
            hamming_dist = options.get("hamming_dist", None)
            if hamming_dist is not None:
                if not (isinstance(hamming_dist, int) and 1 <= hamming_dist <= n_qubits):
                    raise ValueError(
                        f"hamming_dist must be an integer type between 1 and {n_qubits}."
                    )

            confusion_matrix_subspace = normalized_subspace_kron(
                noise_matrices, p_raw.nonzero()[0], hamming_dist
            )

            # GMRES (Generalized minimal residual) for linear equations in higher dimension
            p_corr, exit_code = gmres(confusion_matrix_subspace, p_raw)

            if exit_code != 0:
                logger.warning(f"GMRES did not converge, exited with code {exit_code}")

            # To ensure that we are not working with negative probabilities
            p_corr = mle_solve(p_corr)

        elif optimization_type == ReadOutOptimization.MAJ_VOTE:
            # Majority vote : lets return just that one bit string with all the counts for now
            p_corr = majority_vote(noise_matrices, p_raw)

        else:
            raise NotImplementedError(
                f"Requested method {optimization_type} does not match supported protocols."
            )

        corrected_counts = np.rint(p_corr * n_shots).astype(int)
        # Renormalize if total counts differs from n_shots.
        corrected_counts = renormalize_counts(corrected_counts=corrected_counts, n_shots=n_shots)

        assert (
            corrected_counts.sum() == n_shots
        ), f"Corrected counts sum: {corrected_counts.sum()}, n_shots: {n_shots}"
        corrected_counters.append(
            Counter(
                {bs: count for bs, count in zip(ordered_bitstrings, corrected_counts) if count > 0}
            )
        )
    return corrected_counters


def mitigate(
    model: QuantumModel,
    options: dict,
    noise: NoiseHandler,
    param_values: dict[str, Tensor] = dict(),
) -> list[Counter]:
    if noise.filter(NoiseProtocol.READOUT) is None:
        raise ValueError("A NoiseProtocol.READOUT model must be provided to .mitigate()")
    samples = options.get("samples", None)
    if samples is None:
        n_shots = options.get("n_shots", None)
        if n_shots is None:
            raise ValueError("A n_shots option must be provided.")
        samples = model.sample(param_values, noise=noise, n_shots=n_shots)
    return mitigation_minimization(noise=noise, options=options, samples=samples)



from __future__ import annotations
from qadence_commons import StrEnum


class ReadOutOptimization(StrEnum):
    # Basic inversion and maximum likelihood estimate
    MLE = "mle"
    # Constrained inverse optimization
    CONSTRAINED = "constrained"
    # Matrix free measurement mitigation
    MTHREE = "mthree"
    # Majority voting
    MAJ_VOTE = "majority_vote"



from __future__ import annotations

import importlib
from collections import Counter
from dataclasses import dataclass

from qadence import NoiseHandler, QuantumModel
from torch import Tensor

from qadence_commons import Protocol

PROTOCOL_TO_MODULE = {
    "twirl": "qadence_mitigation.twirl",
    "readout": "qadence_mitigation.readout",
    "zne": "qadence_mitigation.analog_zne",
}


@dataclass
class Mitigations(Protocol):
    TWIRL = "twirl"
    READOUT = "readout"
    ANALOG_ZNE = "zne"

    def __init__(self, protocol: str, options: dict = dict()) -> None:
        super().__init__(protocol, options)

    def __call__(
        self,
        noise: NoiseHandler,
        model: QuantumModel | None = None,
        param_values: dict[str, Tensor] = dict(),
    ) -> list[Counter]:
        try:
            module = importlib.import_module(PROTOCOL_TO_MODULE[self.protocol])
        except (KeyError, ModuleNotFoundError, ImportError) as e:
            raise type(e)(f"Failed to import Mitigations due to {e}.")
        migitation_fn = getattr(module, "mitigate")
        mitigated_counters: list[Counter] = migitation_fn(
            model=model, options=self.options, noise=noise, param_values=param_values
        )
        return mitigated_counters
